""" Tests for system tools """

import os
import subprocess
import pytest
from ubiquerg import is_command_callable
from veracitools import ExpectContext

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"

EXTENSIONS = [".py", ".rb", ".sh", ".java", ".jar", ".pl", ".o", ".R", ".r",
              ".cpp", ".c", ".hs", ".scala", ".class"]


def _mkfile(f):
    """ 'touch' a file. """
    with open(f, 'w'):
        print("Touch: {}".format(f))
    return f


@pytest.mark.parametrize(["cmd", "exp"], [
    ("man", True), ("not-a-cmd", False), (None, TypeError), ("", ValueError)])
def test_command_callability_check_basic(cmd, exp):
    """ Verify expected behavior of command callability checker. """
    with ExpectContext(exp, is_command_callable) as check_callable:
        check_callable(cmd)


@pytest.mark.parametrize(
    "relpath",
    ["subdir", os.path.join("subA", "subB")] +
    ["testfile" + ext for ext in EXTENSIONS])
def test_missing_path_isnt_callable(tmpdir, relpath):
    """ A filepath that doesn't exist (and isn't on PATH) isn't callable. """
    p = os.path.join(tmpdir.strpath, relpath)
    assert not os.path.exists(p)
    assert not is_command_callable(p)


@pytest.mark.parametrize(
    "get_arg", [lambda d: d, lambda d: os.path.join(d, "subdir")])
def test_extant_folder_isnt_callable(tmpdir, get_arg):
    """ A directory path can't be callable. """
    p = get_arg(tmpdir.strpath)
    print("p: {}".format(p))
    if not os.path.exists(p):
        os.makedirs(p)
    assert os.path.exists(p)
    assert not is_command_callable(p)


@pytest.mark.parametrize("filepath", [
    lambda d: os.path.join(d, "testfile" + ext) for ext in EXTENSIONS])
@pytest.mark.parametrize(["setup", "exp_exe"], [
    (lambda _: None, False),
    (lambda f: subprocess.check_call(["chmod", "+x", f]), True)])
def test_extant_file_requires_exec_for_callability(tmpdir, filepath, setup, exp_exe):
    """ Filepath is callable iff it has exec bit. """
    f = filepath(tmpdir.strpath)
    assert not os.path.exists(f)
    _mkfile(f)
    assert os.path.isfile(f)
    setup(f)
    assert os.access(f, os.X_OK) is exp_exe
    assert is_command_callable(f) is exp_exe



""" Functions facilitating file operations """

import os
import sys
import errno
import time

from warnings import warn
from tarfile import open as topen
from hashlib import md5

__all__ = [
    "checksum",
    "size",
    "filesize_to_str",
    "untar",
    "create_lock",
    "remove_lock",
    "wait_for_lock",
    "create_file_racefree",
    "make_lock_path",
]
FILE_SIZE_UNITS = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
LOCK_PREFIX = "lock."


def checksum(path, blocksize=int(2e9)):
    """
    Generate a md5 checksum for the file contents in the provided path.

    :param str path: path to file for which to generate checksum
    :param int blocksize: number of bytes to read per iteration, default: 2GB
    :return str: checksum hash
    """
    m = md5()
    with open(path, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


def size(path, size_str=True):
    """
    Gets the size of a file or directory or list of them in the provided path

    :param str|list path: path or list of paths to the file or directories
        to check size of
    :param bool size_str: whether the size should be converted to a
        human-readable string, e.g. convert B to MB
    :return int|str: file size or file size string
    """

    if isinstance(path, list):
        s_list = sum(filter(None, [size(x, size_str=False) for x in path]))
        return filesize_to_str(s_list) if size_str else s_list

    if os.path.isfile(path):
        s = os.path.getsize(path)
    elif os.path.isdir(path):
        s = 0
        symlinks = []
        for dirpath, dirnames, filenames in os.walk(path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if not os.path.islink(fp):
                    s += os.path.getsize(fp)
                else:
                    s += os.lstat(fp).st_size
                    symlinks.append(fp)
        if len(symlinks) > 0:
            print(
                "{} symlinks were found: {}".format(len(symlinks), "\n".join(symlinks))
            )
    else:
        warn("size could not be determined for: {}".format(path))
        s = None
    return filesize_to_str(s) if size_str else s


def filesize_to_str(size):
    """
    Converts the numeric bytes to the size string

    :param int|float size: file size to convert
    :return str: file size string
    """
    if isinstance(size, (int, float)):
        for unit in FILE_SIZE_UNITS:
            if size < 1024:
                return "{}{}".format(round(size, 1), unit)
            size /= 1024
    warn(
        "size argument was neither an int nor a float, " "returning the original object"
    )
    return size


def untar(src, dst):
    """
    Unpack a path to a target folder.
    All the required directories will be created

    :param str src: path to unpack
    :param str dst: path to output folder
    """
    with topen(src) as tf:
        tf.extractall(path=dst)


def get_file_mod_time(pth):
    """
    Safely get last modification time for a file. Prevents situation when file
    is deleted between file existence check and last file modification check.

    :param str pth: file path to check
    :return float: number of seconds since Jan 1, 1970 00:00:00
    """
    try:
        return os.path.getmtime(pth)
    except Exception as e:
        print(
            "Could not determine timestamp for '{}'. Returning current time. "
            "Caught exception: {}".format(pth, getattr(e, "message", repr(e)))
        )
        return time.time()


def wait_for_lock(lock_file, wait_max=30):
    """
    Just sleep until the lock_file does not exist

    :param str lock_file: Lock file to wait upon.
    :param int wait_max: max wait time if the file in question is already locked
    """
    sleeptime = 0.001
    first_message_flag = False
    dot_count = 0
    totaltime = 0
    ori_timestamp = None
    if os.path.isfile(lock_file):
        ori_timestamp = get_file_mod_time(lock_file)
    while os.path.isfile(lock_file):
        if first_message_flag is False:
            sys.stdout.write("Waiting for file lock: {} ".format(lock_file))
            first_message_flag = True
        else:
            sys.stdout.write(".")
            dot_count += 1
            if dot_count % 60 == 0:
                sys.stdout.write("")
        sys.stdout.flush()
        time.sleep(sleeptime)
        totaltime += sleeptime
        sleeptime = min((sleeptime + 0.1) * 1.25, 10)
        if totaltime >= wait_max:
            if os.path.isfile(lock_file):
                timestamp = get_file_mod_time(lock_file)
                if ori_timestamp and timestamp > ori_timestamp:
                    ori_timestamp = timestamp
                    totaltime = 0
                    sleeptime = 0
                    continue
                raise RuntimeError(
                    "The maximum wait time ({}) has been reached and the lock "
                    "file still exists.".format(wait_max)
                )
    if first_message_flag:
        print(" File unlocked")


def create_file_racefree(file):
    """
    Creates a file, but fails if the file already exists.

    This function will thus only succeed if this process actually creates
    the file; if the file already exists, it will cause an
    OSError, solving race conditions.

    :param str file: File to create.
    :raise OSError: if the file to be created already exists
    """
    write_lock_flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
    fd = os.open(file, write_lock_flags)
    os.close(fd)
    return file


def make_lock_path(lock_name_base):
    """
    Create a collection of path to locks file with given name as bases.

    :param str | list[str] lock_name_base: Lock file names
    :return str | list[str]: Path to the lock files.
    """

    def _mk_lock(lnb):
        base, name = os.path.split(lnb)
        lock_name = name if name.startswith(LOCK_PREFIX) else LOCK_PREFIX + name
        return lock_name if not base else os.path.join(base, lock_name)

    return (
        [_mk_lock(x) for x in lock_name_base]
        if isinstance(lock_name_base, list)
        else _mk_lock(lock_name_base)
    )


def remove_lock(filepath):
    """
    Remove lock

    :param str filepath: path to the file to remove the lock for.
        Not the path to the lock!
    :return bool: whether the lock was found and removed
    """
    lock = make_lock_path(filepath)
    if os.path.exists(lock):
        os.remove(lock)
        return True
    return False


def _create_lock(lock_path, filepath, wait_max):
    try:
        create_file_racefree(lock_path)
    except FileNotFoundError:
        parent_dir = os.path.dirname(filepath)
        os.makedirs(parent_dir)
        _create_lock(lock_path, filepath, wait_max)
    except Exception as e:
        if e.errno == errno.EEXIST:
            # Rare case: file already exists;
            # the lock has been created in the split second since the
            # last lock existence check,
            # wait for the lock file to be gone, but no longer than
            # `wait_max`.
            print(
                "The lock has been created in the split second since the "
                "last lock existence check. Waiting"
            )
            wait_for_lock(lock_path, wait_max)
            _create_lock(lock_path, filepath, wait_max)
        else:
            raise e


def create_lock(filepath, wait_max=10):
    """
    Securely create a lock file

    :param str filepath: path to a file to lock
    :param int wait_max: max wait time if the file in question is already locked
    """
    lock_path = make_lock_path(filepath)
    # wait until no lock is present
    wait_for_lock(lock_path, wait_max)
    _create_lock(lock_path, filepath, wait_max)



""" Tests for web """

import pytest
from ubiquerg.web import is_url


@pytest.mark.parametrize("s", ["https://www.github.com",
                                    "https://www.youtube.com"]
                         )
def test_is_url_tests_positive(s):
    assert is_url(s)


@pytest.mark.parametrize("s", ["www.github.com",
                                    "test: string spaces",
                                    "j%2vv@::https://test.com"]
                         )
def test_is_url_tests_negative(s):
    assert not is_url(s)



""" Tests for checksum """

import hashlib
import os
import itertools
import pytest

from tempfile import mkdtemp
from ubiquerg import checksum, size, filesize_to_str, make_lock_path, \
    create_file_racefree, wait_for_lock, create_lock, remove_lock


def pytest_generate_tests(metafunc):
    """ Dynamic test case generation/parameterization for this module. """
    if "size1" in metafunc.fixturenames and "size2" in metafunc.fixturenames:
        metafunc.parametrize(
            ["size1", "size2"], itertools.product([1, 4], [2, 8]))
    if "lines" in metafunc.fixturenames:
        metafunc.parametrize("lines", [[], ["line1"], ["line1", "line2"]])


def test_checksum(size1, size2, lines, tmpdir):
    """ Checksum result matches expectation and is blocksize-agnostic """
    fp = tmpdir.join("temp-data.txt").strpath
    data = "\n".join(lines)
    with open(fp, 'w') as f:
        f.write(data)
    exp = hashlib.new("md5", data.encode("utf-8")).hexdigest()
    res1 = checksum(fp, size1)
    res2 = checksum(fp, size2)
    assert exp == res1
    assert res1 == res2
    assert res2 == exp


def test_size_returns_str(lines, tmpdir):
    """ Size returns a string and works with both files and directories """
    fp = tmpdir.join("temp-data.txt").strpath
    data = "\n".join(lines)
    with open(fp, 'w') as f:
        f.write(data)
    assert isinstance(size(fp), str)
    assert isinstance(size(tmpdir.strpath), str)


def test_size_returns_int(lines, tmpdir):
    fp = tmpdir.join("temp-data.txt").strpath
    fp_larger = tmpdir.join("temp-data.txt").strpath
    data = "\n".join(lines)
    with open(fp, 'w') as f:
        f.write(data)
    with open(fp_larger, 'w') as f1:
        f1.write(data * 100)
    assert isinstance(size(tmpdir.strpath, False), int)
    assert isinstance(size(fp, False), int)
    assert size(fp, size_str=False) <= size(fp, size_str=False)


def test_nonexistent_path(tmpdir):
    """ Nonexistent path to checksum is erroneous. """
    with pytest.raises(IOError):
        checksum(tmpdir.join("does-not-exist.txt").strpath)


@pytest.mark.parametrize("size_num", list(range(0, 10)) + [i/3 for i in range(0, 10)])
def test_filesize_to_str_int(size_num):
    """ Works with int and returns str """
    assert isinstance(filesize_to_str(size_num), str)


@pytest.mark.parametrize("obj", ["test", [], tuple()])
def test_filesize_to_str_other(obj):
    """ Returns the original object if it's not an int or float and warns """
    with pytest.warns(UserWarning):
        assert filesize_to_str(obj) == obj


class TestLocking:
    @pytest.mark.parametrize("pth", ["test", "", "/home/me/test.yaml", "1"])
    def test_lock_path_creation_string(self, pth):
        assert "lock." in make_lock_path(pth)

    @pytest.mark.parametrize("pth", [["test", ""], ["/home/me/test.yaml", "1"]])
    def test_lock_path_creation_multi(self, pth):
        assert len(make_lock_path(pth)) == len(pth)

    @pytest.mark.parametrize("fn", ["a.yaml", "a.txt"])
    def test_racefree_file_creation_basic(self, fn):
        fp = os.path.join(mkdtemp(), fn)
        create_file_racefree(fp)
        assert os.path.exists(fp)

    @pytest.mark.parametrize("fn", ["a.yaml", "a.txt"])
    def test_racefree_file_creation_errors(self, fn):
        fp = os.path.join(mkdtemp(), fn)
        fh = os.open(fp, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        os.close(fh)
        with pytest.raises(OSError):
            create_file_racefree(fp)

    @pytest.mark.parametrize("lp", ["a.yaml", "a.txt"])
    def test_waiting_for_lock_basic(self, lp):
        wait_for_lock(lp)

    @pytest.mark.parametrize("ln", ["a.yaml", "a.txt"])
    def test_waiting_for_lock_exists(self, ln):
        lp = os.path.join(mkdtemp(), ln)
        fh = os.open(lp, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        os.close(fh)
        with pytest.raises(RuntimeError):
            wait_for_lock(lp, 0.01)

    @pytest.mark.parametrize("fn", ["a.yaml", "a.txt"])
    def test_lock_file_creation_and_removal(self, fn):
        td = mkdtemp()
        create_lock(os.path.join(td, fn))
        locks_list = \
            [os.path.join(td, f) for f in os.listdir(td) if f.startswith("lock.")]
        assert len(locks_list) == 1
        remove_lock(os.path.join(td, fn))
        locks_list = \
            [os.path.join(td, f) for f in os.listdir(td) if f.startswith("lock.")]
        assert len(locks_list) == 0



""" Tools for working with collections """

import itertools
import sys
from warnings import warn

if sys.version_info < (3, 3):
    from collections import Iterable
else:
    from collections.abc import Iterable

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


__all__ = ["is_collection_like", "powerset", "asciify_dict", "merge_dicts"]


def merge_dicts(x, y):
    """
    Merge dictionaries

    :param Mapping x: dict to merge
    :param Mapping y: dict to merge
    :return Mapping: merged dict
    """
    z = x.copy()
    z.update(y)
    return z


def is_collection_like(c):
    """

    Determine whether an object is collection-like.

    :param object c: Object to test as collection
    :return bool: Whether the argument is a (non-string) collection
    """
    return isinstance(c, Iterable) and not isinstance(c, str)


def uniqify(seq):  # Dave Kirby
    """
    Return only unique items in a sequence, preserving order

    :param list seq: List of items to uniqify
    :return list[object]: Original list with duplicates removed
    """
    # Order preserving
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]


def powerset(items, min_items=None, include_full_pop=True, nonempty=False):
    """
    Build the powerset of a collection of items.

    :param Iterable[object] items: "Pool" of all items, the population for
        which to build the power set.
    :param int min_items: Minimum number of individuals from the population
        to allow in any given subset.
    :param bool include_full_pop: Whether to include the full population in
        the powerset (default True to accord with genuine definition)
    :param bool nonempty: force each subset returned to be nonempty
    :return list[object]: Sequence of subsets of the population, in
        nondecreasing size order
    :raise TypeError: if minimum item count is specified but is not an integer
    :raise ValueError: if minimum item count is insufficient to guarantee
        nonempty subsets
    """
    if min_items is None:
        min_items = 1 if nonempty else 0
    else:
        if not isinstance(min_items, int):
            raise TypeError(
                "Min items count for each subset isn't an integer: "
                "{} ({})".format(min_items, type(min_items))
            )
        if nonempty and min_items < 1:
            raise ValueError(
                "When minimum item count is {}, nonempty subsets "
                "cannot be guaranteed.".format(min_items)
            )
    # Account for iterable burn possibility; besides, collection should be
    # relatively small if building the powerset.
    items = list(items)
    n = len(items)
    if n == 0 or n < min_items:
        return []
    max_items = len(items) + 1 if include_full_pop else len(items)
    return list(
        itertools.chain.from_iterable(
            itertools.combinations(items, k) for k in range(min_items, max_items)
        )
    )


def asciify_dict(data):
    """ https://gist.github.com/chris-hailstorm/4989643 """

    def _asciify_list(lst):
        ret = []
        for item in lst:
            if isinstance(item, unicode):
                item = item.encode("utf-8")
            elif isinstance(item, list):
                item = _asciify_list(item)
            elif isinstance(item, dict):
                item = asciify_dict(item)
            ret.append(item)
        return ret

    if sys.version_info[0] >= 3:
        warn("This operation is supported only in Python2", UserWarning)
        return data

    ret = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode("utf-8")
        if isinstance(value, unicode):
            value = value.encode("utf-8")
        elif isinstance(value, list):
            value = _asciify_list(value)
        elif isinstance(value, dict):
            value = asciify_dict(value)
        ret[key] = value
    return ret



""" Tests for rendering CLI options and arguments """

from collections import OrderedDict
import pytest
from ubiquerg import build_cli_extra, powerset, VersionInHelpParser, convert_value
import argparse


def pytest_generate_tests(metafunc):
    """ Test case generation and parameterization for this module """
    if "ordwrap" in metafunc.fixturenames:
        metafunc.parametrize("ordwrap", [tuple, OrderedDict])


def build_parser():
    """
    Example argument parser, needed solely for testing purposes.
    This example parser was copied from looper/__init__.py

    :return argparse.ArgumentParser
    """

    banner = "%(prog)s - Loop through samples and submit pipelines."
    additional_description = "For subcommand-specific options, " \
                             "type: '%(prog)s <subcommand> -h'"
    additional_description += "\nhttps://github.com/pepkit/looper"

    parser = VersionInHelpParser(prog="looper", description=banner,
        epilog=additional_description, version=0.1)

    # Logging control
    parser.add_argument("--logfile", dest="logfile",
        help="Optional output file for looper logs (default: %(default)s)")
    parser.add_argument("--verbosity", dest="verbosity", type=int,
        choices=range(4),
        help="Choose level of verbosity (default: %(default)s)")
    parser.add_argument("--logging-level", dest="logging_level",
        help=argparse.SUPPRESS)
    parser.add_argument("--dbg", dest="dbg", action="store_true",
        help="Turn on debug mode (default: %(default)s)")
    parser.add_argument("--env", dest="env", default=None,
        help="Environment variable that points to the DIVCFG file. "
             "(default: DIVCFG)")
    parser.add_argument("--dotfile-template", action="store_true",
                        help="Print out a looper dotfile template and exit")

    # Individual subcommands
    msg_by_cmd = {"run": "Main Looper function: Submit jobs for samples.",
        "rerun": "Resubmit jobs with failed flags.",
        "runp": "Submit jobs for a project.",
        "summarize": "Summarize statistics of project samples.",
        "destroy": "Remove all files of the project.",
        "check": "Checks flag status of current runs.",
        "clean": "Runs clean scripts to remove intermediate "
                 "files of already processed jobs."}

    subparsers = parser.add_subparsers(dest="command")

    def add_subparser(cmd):
        message = msg_by_cmd[cmd]
        return subparsers.add_parser(cmd, description=message, help=message)

    # Run and rerun command
    run_subparser = add_subparser("run")
    rerun_subparser = add_subparser("rerun")
    collate_subparser = add_subparser("runp")
    for subparser in [run_subparser, rerun_subparser, collate_subparser]:
        subparser.add_argument("--ignore-flags", dest="ignore_flags",
            default=False, action="store_true",
            help="Ignore run status flags? Default: False. "
                 "By default, pipelines will not be submitted if a pypiper "
                 "flag file exists marking the run (e.g. as "
                 "'running' or 'failed'). Set this option to ignore flags "
                 "and submit the runs anyway. Default=False")
        subparser.add_argument("-t", "--time-delay", dest="time_delay",
            type=int, default=0,
            help="Time delay in seconds between job submissions.")
        subparser.add_argument("-p", "--package",
            help="Name of computing resource package to use")
        subparser.add_argument("--compute",
            help="Specification of individual computing resource settings; "
                 "separate setting name/key from value with equals sign, "
                 "and separate key-value pairs from each other by comma")
        subparser.add_argument("--limit", dest="limit", default=None,
            type=int, help="Limit to n samples.")
        subparser.add_argument("-a", "--pipeline-args",
            dest="pipeline_args", default="",
            help="arguments to pass to a pipline")
        subparser.add_argument("-s", "--settings", dest="settings",
            default="",
            help="path to a YAML-formatted settings file used to populate "
                 "the command template")
    for subparser in [run_subparser, rerun_subparser]:
        # Note that defaults for otherwise numeric lump parameters are
        # set to
        # null by default so that the logic that parses their values may
        # distinguish between explicit 0 and lack of specification.
        subparser.add_argument("--lump", default=None,
            type=float,
            help="Maximum total input file size for a lump/batch of "
                 "commands in a single job (in GB)")
        subparser.add_argument("--lumpn", default=None,
            type=int,
            help="Number of individual scripts grouped into "
                 "single submission")

    # Other commands
    summarize_subparser = add_subparser("summarize")
    destroy_subparser = add_subparser("destroy")
    check_subparser = add_subparser("check")
    clean_subparser = add_subparser("clean")

    check_subparser.add_argument("-A", "--all-folders",
        action="store_true", default=False,
        help="Check status for all project's output folders, not just "
             "those for samples specified in the config file used. "
             "Default=False")

    for subparser in [destroy_subparser, clean_subparser]:
        subparser.add_argument("--force-yes", action="store_true",
            default=False,
            help="Provide upfront confirmation of destruction intent, "
                 "to skip console query.  Default=False")

    # Common arguments
    for subparser in [run_subparser, rerun_subparser, summarize_subparser,
                      destroy_subparser, check_subparser, clean_subparser,
                      collate_subparser]:
        subparser.add_argument("config_file", nargs="?",
                               help="Project configuration file (YAML).")
        # subparser.add_argument(
        #         "-c", "--config", required=False, default=None,
        #         dest="looper_config", help="Looper configuration file (
        #         YAML).")
        subparser.add_argument("--pipeline-interfaces", dest="pifaces",
            nargs="+", action='append',
            help="Path to a pipeline interface file")
        subparser.add_argument("--file-checks", dest="file_checks",
            action="store_true",
            help="Perform input file checks. Default=True.")
        subparser.add_argument("-d", "--dry-run", dest="dry_run",
            action="store_true",
            help="Don't actually submit the jobs.  Default=False")

        fetch_samples_group = subparser.add_argument_group("select samples",
            "This group of arguments lets you specify samples to use by "
            "exclusion OR inclusion of the samples attribute values.")
        fetch_samples_group.add_argument("--selector-attribute",
            dest="selector_attribute", default="toggle",
            help="Specify the attribute for samples exclusion OR inclusion")
        protocols = fetch_samples_group.add_mutually_exclusive_group()
        protocols.add_argument("--selector-exclude", nargs='*',
            dest="selector_exclude",
            help="Operate only on samples that either lack this attribute "
                 "value or for which this value is not in this collection.")
        protocols.add_argument("--selector-include", nargs='*',
            dest="selector_include",
            help="Operate only on samples associated with these attribute "
                 "values; if not provided, all samples are used.")
        subparser.add_argument("--amendments", dest="amendments", nargs="+",
            help="Name of amendment(s) to use, as designated in the "
                 "project's configuration file")

    return parser, msg_by_cmd


@pytest.mark.parametrize(["optargs", "expected"], [
    ([("-X", None), ("--revert", 1), ("-O", "outfile"),
      ("--execute-locally", None), ("-I", ["file1", "file2"])],
     "-X --revert 1 -O outfile --execute-locally -I file1 file2")
])
def test_build_cli_extra(optargs, expected, ordwrap):
    """ Check that CLI optargs are rendered as expected. """
    observed = build_cli_extra(ordwrap(optargs))
    print("expected: {}".format(expected))
    print("observed: {}".format(observed))
    assert expected == observed


@pytest.mark.parametrize(
    "optargs", powerset([(None, "a"), (1, "one")], nonempty=True))
def test_illegal_cli_extra_input_is_exceptional(optargs, ordwrap):
    """ Non-string keys are illegal and cause a TypeError. """
    with pytest.raises(TypeError):
        build_cli_extra(ordwrap(optargs))


def test_dests_by_subparser_return_type():
    """ Check if the return type is dict of lists keyed by subcommand name """
    parser, msgs_by_cmd = build_parser()
    assert isinstance(parser.dests_by_subparser(), dict)
    assert all([isinstance(v, list)
                for k, v in parser.dests_by_subparser().items()])
    assert all([k in parser.dests_by_subparser()
                for k in list(msgs_by_cmd.keys())])

@pytest.mark.parametrize("subcmd",
                         list(build_parser()[0].dests_by_subparser().keys()))
def test_dests_by_subparser_specific_subcmd(subcmd):
    """
    Should return only specified section of the
    return if subparser specified
    """
    parser, _ = build_parser()
    assert isinstance(parser.dests_by_subparser(subcmd), dict)
    assert subcmd in parser.dests_by_subparser(subcmd).keys()
    assert isinstance(parser.dests_by_subparser(subcmd)[subcmd], list)


def test_dests_by_subparser_toplevel_actions():
    """
    Should return only top level dests, which is a list
    """
    parser, _ = build_parser()
    tld = parser.dests_by_subparser(top_level=True)
    assert isinstance(tld, list)
    assert len(tld) == 6


def test_arg_defaults_return_type():
    """ Check if the return type is dict of dicts keyed by subcommand name """
    parser, msgs_by_cmd = build_parser()
    assert isinstance(parser.arg_defaults(), dict)
    assert all([isinstance(v, dict)
                for k, v in parser.arg_defaults().items()])
    assert all([k in parser.arg_defaults()
                for k in list(msgs_by_cmd.keys())])


@pytest.mark.parametrize("subcmd",
                         list(build_parser()[0].dests_by_subparser().keys()))
def test_arg_defaults_specific_subcmd(subcmd):
    """
    Should return only specified section of the
    return if subparser specified
    """
    parser, _ = build_parser()
    assert isinstance(parser.arg_defaults(subcmd), dict)
    assert subcmd in parser.arg_defaults(subcmd).keys()
    assert isinstance(parser.arg_defaults(subcmd)[subcmd], dict)


def test_arg_defaults_toplevel_actions():
    """
    Should return only top level defaults, which is a dict
    """
    parser, _ = build_parser()
    tld = parser.arg_defaults(top_level=True)
    assert isinstance(tld, dict)
    assert len(tld) == 6


def test_arg_defaults_unqiue():
    """
    Should return only dict of unique kv pairs, not keyed by subcommands
    """
    parser, _ = build_parser()
    tld = parser.arg_defaults(unique=True)
    assert isinstance(tld, dict)
    assert len(tld) == 17
    assert len(set(tld.keys())) == len(tld.keys())


@pytest.mark.parametrize(["input", "output"],
                         [(None, None), ("None", None), ("1", 1), ("0.1", 0.1),
                          ("True", True), ("False", False)])
def test_value_conversion(input, output):
    assert convert_value(input) == output



""" Tests for environment-related functionality """

import os
import pytest
from ubiquerg import TmpEnv


@pytest.mark.parametrize("envvars", [
    {"DUMMYTESTVAR": "arbitrary"},
    {"DUMMYTESTVAR": "random", "TOTALLY_RAD_VARIABLE": "ephemeral"}
])
def test_overwrite_prohibited(envvars):
    """ Prohibition of overwrite causes ValueError in case of collision. """
    check_unset(envvars)
    with TmpEnv(**envvars), pytest.raises(ValueError):
        TmpEnv(overwrite=False, **{k: v + "_EXTRA" for k, v in envvars.items()})
    check_unset(envvars)


@pytest.mark.parametrize("envvars", [
    {"DUMMYTESTVAR": "arbitrary"},
    {"DUMMYTESTVAR": "random", "TOTALLY_RAD_VARIABLE": "ephemeral"}
])
def test_overwrite_allowed(envvars):
    """ Verify that overwriting existing value with different one works if permitted. """
    check_unset(envvars)
    with TmpEnv(**envvars):
        modified = {k: v + "_EXTRA" for k, v in envvars.items()}
        with TmpEnv(overwrite=True, **modified):
            for k, v in modified.items():
                assert v == os.getenv(k)
    check_unset(envvars)


@pytest.mark.parametrize("overwrite", [False, True])
@pytest.mark.parametrize("envvars", [
    {"DUMMYTESTVAR": "arbitrary"},
    {"DUMMYTESTVAR": "random", "TOTALLY_RAD_VARIABLE": "ephemeral"}
])
def test_no_intersection(overwrite, envvars):
    """ Verify that overwrite permission is irrelevant if no variable name collision occurs. """
    check_unset(envvars)
    with TmpEnv(overwrite=overwrite, **envvars):
        for k, v in envvars.items():
            assert v == os.getenv(k)


def check_unset(envvars):
    """ Verify that each environment variable is not set. """
    fails = [v for v in envvars if os.getenv(v) or v in os.environ]
    assert not fails, "Vars set: {}".format(fails)



""" Functions for working with command-line interaction """

from .collection import is_collection_like, merge_dicts
from argparse import (
    ArgumentParser,
    _SubParsersAction,
    _HelpAction,
    _VersionAction,
    SUPPRESS,
)
import sys

__classes__ = ["VersionInHelpParser"]
__all__ = __classes__ + ["build_cli_extra", "query_yes_no", "convert_value"]


class VersionInHelpParser(ArgumentParser):
    def __init__(self, version=None, **kwargs):
        """Overwrites the inherited init. Saves the version as an object
        attribute for further use."""
        super(VersionInHelpParser, self).__init__(**kwargs)
        self.version = version
        if self.version is not None:
            self.add_argument(
                "--version",
                action="version",
                version="%(prog)s {}".format(self.version),
            )

    def format_help(self):
        """ Add version information to help text. """
        help_string = (
            "version: {}\n".format(str(self.version))
            if self.version is not None
            else ""
        )
        return help_string + super(VersionInHelpParser, self).format_help()

    def subparsers(self):
        """
        Get the subparser associated with a parser.

        :return argparse._SubparsersAction: action defining the subparsers
        """
        subs = [a for a in self._actions if isinstance(a, _SubParsersAction)]
        if len(subs) != 1:
            raise ValueError("Expected exactly 1 subparser, got {}".format(len(subs)))
        return subs[0]

    def top_level_args(self):
        """
        Get actions not assiated with any subparser.

        Help and version are also excluded

        :return list[argparse.<action_type>]: list of argument actions
        """
        excl = [_SubParsersAction, _HelpAction, _VersionAction]
        return [a for a in self._actions if not type(a) in excl]

    def subcommands(self):
        """
        Get subcommands defined by a parser.

        :return list[str]: subcommands defined within this parser
        """
        return list(self.subparsers().choices.keys())

    def dests_by_subparser(self, subcommand=None, top_level=False):
        """
        Get argument dests by subcommand from a parser.

        :param str subcommand: subcommand to get dests for
        :return dict: dests by subcommand
        """
        if top_level:
            top_level_actions = self.top_level_args()
            dest_list = []
            for tla in top_level_actions:
                if hasattr(tla, "dest"):
                    dest_list.append(tla.dest)
            return dest_list

        if subcommand is not None and subcommand not in self.subcommands():
            raise ValueError(
                "'{}' not found in this parser commands: {}".format(
                    subcommand, str(self.subcommands())
                )
            )
        subs = (
            self.subparsers().choices
            if subcommand is None
            else {subcommand: self.subparsers().choices[subcommand]}
        )
        dests = {}
        for subcmd, sub in subs.items():
            dest_list = []
            for action in sub._actions:
                if isinstance(action, _HelpAction):
                    continue
                if hasattr(action, "dest"):
                    dest_list.append(action.dest)
            dests[subcmd] = dest_list
        return dests

    def suppress_defaults(self):
        """
        Remove parser change defaults to argparse.SUPPRESS so that they do not
        show up in the argparse.Namespace object after argument parsing.
        """
        top_level_actions = self.top_level_args()
        for tla in top_level_actions:
            if hasattr(tla, "dest"):
                tla.dest = SUPPRESS
        subs = self.subparsers().choices
        for subcmd, sub in subs.items():
            for sa in sub._actions:
                if hasattr(sa, "dest"):
                    sa.default = SUPPRESS

    def arg_defaults(self, subcommand=None, unique=False, top_level=False):
        """
        Get argument defaults by subcommand from a parser.

        :param str subcommand: subcommand to get defaults for
        :param bool unique: whether only unique flat dict of dests and
            defaults mapping should be returned
        :return dict: defaults by subcommand
        """
        if top_level:
            top_level_actions = self.top_level_args()
            defaults_dict = {}
            for tla in top_level_actions:
                if hasattr(tla, "default") and hasattr(tla, "dest"):
                    defaults_dict.update({tla.dest: tla.default})
            return defaults_dict

        if subcommand is not None and subcommand not in self.subcommands():
            raise ValueError(
                "'{}' not found in this parser commands: {}".format(
                    subcommand, str(self.subcommands())
                )
            )
        subs = (
            self.subparsers().choices
            if subcommand is None
            else {subcommand: self.subparsers().choices[subcommand]}
        )
        defaults = {}
        for subcmd, sub in subs.items():
            defaults_dict = {}
            for action in sub._actions:
                if isinstance(action, _HelpAction):
                    continue
                if hasattr(action, "default") and hasattr(action, "dest"):
                    defaults_dict.update({action.dest: action.default})
            defaults[subcmd] = defaults_dict
            if unique:
                unique_defaults = {}
                for k, v in defaults.items():
                    unique_defaults = merge_dicts(unique_defaults, v)
                return unique_defaults
        return defaults


def build_cli_extra(optargs):
    """
    Render CLI options/args as text to add to base command.

    To specify a flag, map an option to None. Otherwise, map option short or
    long name to value(s). Values that are collection types will be rendered
    with single space between each. All non-string values are converted to
    string.

    :param Mapping | Iterable[(str, object)] optargs: values used as
        options/arguments
    :return str: text to add to base command, based on given opts/args
    :raise TypeError: if an option name isn't a string
    """

    def render(k, v):
        if not isinstance(k, str):
            raise TypeError("Option name isn't a string: {} ({})".format(k, type(k)))
        if v is None:
            return k
        if is_collection_like(v):
            v = " ".join(map(str, v))
        return "{} {}".format(k, v)

    try:
        data_iter = optargs.items()
    except AttributeError:
        data_iter = optargs

    return " ".join(render(*kv) for kv in data_iter)


def query_yes_no(question, default="no"):
    """
    Ask a yes/no question via raw_input() and return their answer.

    :param str question: a string that is presented to the user.
    :param str default: the presumed answer if the user just hits <Enter>.
    :return bool: True for "yes" or False for "no"
    """

    def parse(ans):
        return {"yes": True, "y": True, "ye": True, "no": False, "n": False}[
            ans.lower()
        ]

    try:
        prompt = {None: "[y/n]", "yes": "[Y/n]", "no": "[y/N]"}[
            None if default is None else default.lower()
        ]
    except (AttributeError, KeyError):
        raise ValueError("invalid default answer: {}".format(default))
    msg = "{q} {p} ".format(q=question, p=prompt)
    while True:
        sys.stdout.write(msg)
        try:
            return parse(_read_from_user() or default)
        except KeyError:
            sys.stdout.write("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")


def convert_value(val):
    """
    Convert string to the most appropriate type, one of:
    bool, str, int, None or float

    :param str val: the string to convert
    :return bool | str | int | float | None: converted string to the
        most appropriate type
    """
    if not isinstance(val, str):
        try:
            val = str(val)
        except:
            raise ValueError(
                "The input has to be of type convertible to 'str',"
                " got '{}'".format(type(val))
            )

    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        if val == "None":
            return None
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False
        try:
            float(val)
        except ValueError:
            return val
        else:
            try:
                int(val)
            except ValueError:
                return float(val)
            else:
                return int(val)


def _read_from_user():
    import sys

    if sys.version_info.major < 3:
        from __builtin__ import raw_input

        return raw_input()
    return input()



""" Tests for collection utilities """

from collections import Counter, OrderedDict
import itertools
import sys
if sys.version_info.major < 3:
    from inspect import getargspec as get_fun_sig
else:
    from inspect import getfullargspec as get_fun_sig
import random
import string
import sys
import pytest
from ubiquerg.collection import *

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


MIN_SIZE_KWARG = "min_size"


def combo_space(ks, items):
    """
    Create the Cartesian product of values sets, each bound to a key.

    :param Iterable[str] ks: subset of keys from the mapping items
    :param Mapping[str, Iterable[object]] items: bindings between key and
        collection of values
    :return itertools.product: Cartesian product of the values sets bound by
        the given subset of keys
    """
    return itertools.product(*[items[k] for k in ks])


def get_default_parameters(func, pred=None):
    """
    For given function, get mapping from parameter name to default value.

    :param callable func: the function to inspect
    :param func(object, object) -> bool pred: how to determine whether the
        parameter should be included, based on name and default value
    :return OrderedDict[str, object]]: mapping from parameter name to default value
    """
    if not callable(func):
        raise TypeError("Not a callable: {} ({})".format(func.__name__, type(func)))
    spec = get_fun_sig(func)
    par_arg_pairs = zip(spec.args[(len(spec.args) - len(spec.defaults)):],
                        spec.defaults)
    return OrderedDict(
        par_arg_pairs if pred is None else
        [(p, a) for p, a in par_arg_pairs if pred(p, a)])


def _generate_too_few_items(minsize, maxsize, pool):
    """
    Generate a "pool" of items of random size guaranteed less than given bound.

    :param int minsize: minimum pool size
    :param int maxsize: maximum pool size
    :param Iterable[object] pool: items from which to pull a subset for the
        returned pool
    :return Iterable[object]: subset of initial pool, constrained to be
        sufficiently small in accordance with given bounds
    :raise TypeError: if either size bound is not an integer
    :raise ValueError: if maxsize < minsize
    """
    print_bound = lambda: "[{}, {}]".format(minsize, maxsize)
    if not (isinstance(minsize, int) and isinstance(maxsize, int)):
        raise TypeError("Size bounds must be integers; got {}".
                        format(print_bound()))
    if maxsize < minsize:
        raise ValueError("Nonesense size bounds: {}".format(print_bound()))
    n = random.randint(minsize, maxsize)
    return n, random.sample(pool, max(n - 1, 0))


POWERSET_BOOL_KWARGSPACE = {
    p: [False, True] for p in
    get_default_parameters(powerset, lambda _, v: isinstance(v, bool))}
KWARGSPACE_POWERSET = {
    ks: combo_space(ks, POWERSET_BOOL_KWARGSPACE)
    for n in range(len(POWERSET_BOOL_KWARGSPACE) + 1)
    for ks in [tuple(c) for c in
               itertools.combinations(POWERSET_BOOL_KWARGSPACE.keys(), n)]}


def pytest_generate_tests(metafunc):
    """ Test case generation and parameterization for this module """
    if "arbwrap" in metafunc.fixturenames:
        metafunc.parametrize("arbwrap", [list, tuple, set, iter])


def randcoll(pool, dt):
    """
    Generate random collection of 1-10 elements.

    :param Iterable pool: elements from which to choose
    :param type dt: type of collection to create
    :return Iterable[object]: collection of randomly generated elements
    """
    valid_types = [tuple, list, set, dict]
    if dt not in valid_types:
        raise TypeError("{} is an invalid type; choose from {}".
                        format(str(dt), ", ".join(str(t) for t in valid_types)))
    rs = [random.choice(pool) for _ in range(random.randint(1, 10))]
    return dict(enumerate(rs)) if dt == dict else rs


@pytest.mark.parametrize(
    ["arg", "exp"],
    [(random.randint(-sys.maxsize - 1, sys.maxsize), False),
     (random.random(), False),
     (random.choice(string.ascii_letters), False),
     ([], True), (set(), True), (dict(), True), (tuple(), True),
     (randcoll(string.ascii_letters, list), True),
     (randcoll(string.ascii_letters, dict), True),
     (randcoll([int(d) for d in string.digits], tuple), True),
     (randcoll([int(d) for d in string.digits], set), True),
     ("", False), ("abc", False)]
)
def test_is_collection_like(arg, exp):
    """ Test arbiter of whether an object is collection-like. """
    assert exp == is_collection_like(arg)


@pytest.mark.parametrize("kwargs",
    [d for ks, vals_list in KWARGSPACE_POWERSET.items()
     for d in [dict(zip(ks, vs)) for vs in vals_list]])
def test_powerset_of_empty_pool(arbwrap, kwargs):
    """ Empty collection's powerset is always empty. """
    assert [] == powerset(arbwrap([]), **kwargs)


@pytest.mark.parametrize("include_full_pop", [False, True])
@pytest.mark.parametrize(
    ["min_items", "pool"],
    [_generate_too_few_items(
        2, 10, list(string.ascii_letters) + list(range(-10, 11)))
     for _ in range(5)])
def test_powerset_fewer_items_than_min(arbwrap, min_items, pool, include_full_pop):
    """ Minimum item count in excess of pool size results in empty powerset. """
    print("pool (n={}): {}".format(len(pool), pool))
    assert [] == powerset(arbwrap(pool), min_items=min_items,
                          include_full_pop=include_full_pop)


@pytest.mark.parametrize("pool", [[1, 2, 3], ["a", "b", "c"]])
def test_full_powerset(arbwrap, pool):
    obs = powerset(arbwrap(pool))
    reps = list(filter(lambda kv: kv[1] > 1, Counter(obs).items()))
    assert [] == reps, "Repeated items in powerset: {}".format(reps)
    assert 2 ** len(pool) == len(obs)


@pytest.mark.parametrize(
    ["pool", "kwargs", "expected"], [
        ([-1, 1], {}, [set(), {-1}, {1}, {-1, 1}]),
        ([-1, 1], {"include_full_pop": False}, [set(), {-1}, {1}]),
        ([-1, 1], {"nonempty": True}, [{-1}, {1}, {-1, 1}]),
        ([-1, 1], {"include_full_pop": False, "nonempty": True}, [{-1}, {1}])
    ])
def test_powerset_legitimate_input(arbwrap, pool, kwargs, expected):
    """ Powerset behavior responds to arguments to its parameters """

    # Verification logic type dependencies
    assert isinstance(expected, list)
    assert all(isinstance(ss, set) for ss in expected)

    observed = powerset(arbwrap(pool), **kwargs)    # Find the powerset.

    # Check for repeats.
    reps = list(filter(lambda kv: kv[1] > 1, Counter(observed).items()))
    assert [] == reps, "Repeated items in powerset: {}".format(reps)
    assert len(expected) == len(observed)    # Dumbest check is on raw size.

    # Better, harder check is on mutual inclusion.
    # With no repeats in the empirical powerset and expected size validated
    # as equal to observe size, to verify bijection we check for all "hits."
    exp_hits = [False] * len(expected)
    for sub in observed:
        sought = set(sub)
        for i, exp in enumerate(expected):
            if sought == exp:
                exp_hits[i] = True
                break
    missed = [ss for ss, hit in zip(expected, exp_hits) if not hit]
    assert [] == missed, "{} missed subsets: {}".format(len(missed), missed)


@pytest.mark.parametrize(
    ["kwargs", "exp_err"],
    [({"min_items": 0, "nonempty": True}, ValueError),
     ({"min_items": -1, "nonempty": True}, ValueError),
     ({"min_items": -sys.maxsize, "nonempty": True}, ValueError),
     ({"min_items": "1"}, TypeError),
     ({"min_items": "None"}, TypeError),
     ({"min_items": [2]}, TypeError)])
@pytest.mark.parametrize("pool", [string.ascii_letters, range(-10, 11)])
def test_powerset_illegal_input(arbwrap, kwargs, exp_err, pool):
    """ Invalid argument combination to powerset parameters is exceptional. """
    invalid_kwargs = set(kwargs.keys()) - set(get_fun_sig(powerset).args)
    assert not invalid_kwargs, \
        "Attempting to use keyword arguments not in the signature of {}: {}".\
        format(powerset.__name__, ", ".join(invalid_kwargs))
    with pytest.raises(exp_err):
        powerset(arbwrap(pool), **kwargs)


@pytest.mark.parametrize("dict", [{"a": 1}, {"b": None}, {"b": [u'test', ["test", u'test']], "c": [{"b": 1}]}])
def test_asciifying_dicts(dict):
    if sys.version_info[0] >= 3:
        with pytest.warns(UserWarning):
            assert dict == asciify_dict(dict)
    else:
        asciify_dict(dict)

@pytest.mark.parametrize(["x", "y"], [({"a": 1}, {"b": 2}), ({"a": 1, "c": 2}, {"b": 2})])
def test_merging_dicts(x, y):
    assert "a" in list(merge_dicts(x, y).keys())



""" Tests for binary user terminal interaction """

import itertools
import mock

import pytest
from ubiquerg import query_yes_no


READ_INPUT_PATH = "ubiquerg.cli_tools._read_from_user"


def pytest_generate_tests(metafunc):
    """ Dynamic test case generation and parameterization for this module. """
    if "question" in metafunc.fixturenames:
        metafunc.parametrize("question", ["want to test?", "will this work?"])


@pytest.mark.parametrize(
    "default", [1, "a", "wontwork", [], {}, (), "Y", "N", "y", "n"])
def test_illegal_default_yields_value_error(question, default):
    """ Illegal default response causes ValueError. """
    with pytest.raises(ValueError):
        query_yes_no(question, default)


@pytest.mark.parametrize(["default", "expected"], [("no", False), ("yes", True)])
def test_query_yesno_empty_with_default(question, default, expected, capsys):
    """ Default response is used when user input is empty. """
    with mock.patch(READ_INPUT_PATH, return_value=""):
        assert expected is query_yes_no(question, default)


@pytest.mark.parametrize(["default", "responses", "expected"], [
    (None, ["noeffect", "N"], False),
    (None, ["meaningless", "Y"], True),
    ("no", ["illegal", "yes"], True),
    ("yes", ["invalid", "no"], False)])
def test_response_sequence(question, default, responses, expected):
    """ The interaction re-prompts and then responds as intended. """
    with mock.patch(READ_INPUT_PATH, side_effect=responses):
        assert expected is query_yes_no(question, default)


@pytest.mark.parametrize("default", [None, "yes", "no", "YES", "NO"])
@pytest.mark.parametrize(["response", "expected"], itertools.chain(*[
    [(ans.lower(), flag), (ans.upper(), flag)] for flag, resps in
    {True: ("yes", "ye", "y"), False: ("no", "n")}.items() for ans in resps]))
def test_query_yesno_input(question, default, response, expected):
    """ Yes/No interaction responds to user input. """
    with mock.patch(READ_INPUT_PATH, return_value=response):
        assert query_yes_no(question, default) is expected



""" Validate what's available directly on the top-level import. """

import pytest
from inspect import isclass, isfunction

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


@pytest.mark.parametrize(
    ["obj_name", "typecheck"],
    [("build_cli_extra", isfunction), ("checksum", isfunction), ("size", isfunction),
     ("expandpath", isfunction), ("is_collection_like", isfunction),
     ("is_command_callable", isfunction), ("is_url", isfunction),
     ("powerset", isfunction), ("query_yes_no", isfunction),
     ("TmpEnv", isclass)])
def test_top_level_exports(obj_name, typecheck):
    """ At package level, validate object availability and type. """
    import ubiquerg
    try:
        obj = getattr(ubiquerg, obj_name)
    except AttributeError:
        pytest.fail("Unavailable on {}: {}".format(ubiquerg.__name__, obj_name))
    else:
        assert typecheck(obj)



""" Tests for filesystem utilities """

import os
import pytest
from ubiquerg import expandpath, parse_registry_path

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"


class PathTestCase(object):
    """ Bundle data and expectation with context management for a test case. """

    def __init__(self, path, expected, **envs):
        """
        Create the case with raw path, expected result, and envvar updates.

        :param str path: raw path value to use as input to function under test
        :param str expected: expected result of applying function under test
            to raw path value
        """
        self.path = path
        self.expected = expected
        self._originals = {k: os.environ[k] for k in envs if k in os.environ}
        self._updates = envs

    def __enter__(self):
        """ Update environment variables. """
        for k, v in self._updates.items():
            os.environ[k] = v
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Restore environment variables. """
        for k in self._updates:
            try:
                os.environ[k] = self._originals[k]
            except KeyError:
                del os.environ[k]
        for k, v in self._originals.items():
            assert v == os.environ[k]
        for k in set(self._updates.keys()) - set(self._originals.keys()):
            assert k not in os.environ

    def __repr__(self):
        """ Conveniently represent the test case instance. """
        return "Path={}, Exp={}, Envs={}".\
            format(self.path, self.expected, self._updates)


@pytest.mark.parametrize("case", [
    PathTestCase(os.path.join("random", "$LIKELY_NEWVAL", "leaf.md"),
                 os.path.join("random", "newval", "leaf.md"),
                 LIKELY_NEWVAL="newval"),
    PathTestCase(os.path.join("arbitrary", "$HOME", "leaf.rst"),
                 os.path.join("arbitrary", "uibvreal", "leaf.rst"),
                 HOME="uibvreal")
])
def test_expandpath_envvars(case):
    """ Validate expected path expansion behavior. """
    with case:
        assert case.expected == expandpath(case.path)


@pytest.mark.xfail(reason="Double slash handling is up for discussion, esp w/ Py3")
@pytest.mark.parametrize("case", [
    PathTestCase("/a/b//c.txt", "/a/b/c.txt"),
    PathTestCase("double//slash//2.txt", "double/slash/2.txt")])
def test_expandpath_doubleslash(case):
    assert case.expected == expandpath(case.path)


def test_parse_reg():
    pvars = parse_registry_path("abc")
    assert(pvars['item'] == 'abc')
    assert(parse_registry_path("http://big.databio.org/bulker/bulker/demo.yaml") == None)



""" Web-related utilities """

import re

__all__ = ["is_url"]


def is_url(maybe_url):
    """
    Determine whether a path is a URL.

    :param str maybe_url: path to investigate as URL
    :return bool: whether path appears to be a URL
    """
    # from Django 1.3.x
    # https://github.com/django/django/blob/6726d750979a7c29e0dd866b4ea367eef7c8a420/django/core/validators.py#L45-L51
    regex = re.compile(
        r"^(?:http|ftp)s?://"
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        r"localhost|"
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
        r"(?::\d+)?"
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    return re.match(regex, maybe_url) is not None
